var Rx = require("rx/dist/rx") && require("rx/dist/rx.aggregates");
var Observable = Rx.Observable;

var GetRequest = require("./../request/GetRequest");
var SetRequest = require("./../request/SetRequest");

var prefix = require("./../internal/prefix");
var getType = require("./../support/get-type");
var isObject = require("./../support/is-object");
var arrayClone = require("./../support/array-clone");
var falcorPathUtils = require("falcor-path-utils");

/* eslint-disable no-labels block-scoped-var */
function RequestQueue(model, scheduler) {
    this.total = 0;
    this.model = model;
    this.requests = [];
    this.scheduler = scheduler;
}

RequestQueue.prototype.get = function getRequest(paths) {

    var self = this;

    return Observable.defer(function() {

        var requests = self.distributePaths(paths, self.requests, GetRequest);

        return (Observable.defer(function() {
                return Observable.fromArray(requests.map(function(request) {
                    return request.getSourceObservable();
                }));
            })
            .mergeAll()
            .reduce(self.mergeJSONGraphs, {
                index: -1,
                jsonGraph: {}
            })
            .map(function(response) {
                return {
                    paths: paths,
                    index: response.index,
                    jsonGraph: response.jsonGraph
                };
            })
            .subscribeOn(self.scheduler).finally(function() {
                var paths2 = arrayClone(paths);
                var pathCount = paths2.length;
                var requestIndex = -1;
                var requestCount = requests.length;
                while (pathCount > 0 && requestCount > 0 && ++requestIndex < requestCount) {
                    var request = requests[requestIndex];
                    if (request.pending) {
                        continue;
                    }
                    var pathIndex = -1;
                    while (++pathIndex < pathCount) {
                        var path = paths2[pathIndex];
                        if (request.removePath(path)) {
                            paths2.splice(pathIndex--, 1);
                            if (--pathCount === 0) {
                                break;
                            }
                        }
                    }
                    if (request.length === 0) {
                        requests.splice(requestIndex--, 1);
                        if (--requestCount === 0) {
                            break;
                        }
                    }
                }
            }));
    });
};

RequestQueue.prototype.set = function setRequest(jsonGraphEnvelope) {
    jsonGraphEnvelope.paths = falcorPathUtils.collapse(jsonGraphEnvelope.paths);
    return SetRequest.create(this.model, jsonGraphEnvelope);
};

RequestQueue.prototype._remove = function removeRequest(request) {
    var requests = this.requests;
    var index = requests.indexOf(request);
    if (index !== -1) {
        requests.splice(index, 1);
    }
};

RequestQueue.prototype.distributePaths = function distributePathsAcrossRequests(paths, requests, RequestType) {

    var model = this.model;
    var pathsIndex = -1;
    var pathsCount = paths.length;

    var requestIndex = -1;
    var requestCount = requests.length;
    var participatingRequests = [];
    var pendingRequest;
    var request;

    insertPath: while (++pathsIndex < pathsCount) {

        var path = paths[pathsIndex];

        requestIndex = -1;

        while (++requestIndex < requestCount) {
            request = requests[requestIndex];
            if (request.insertPath(path, request.pending)) {
                participatingRequests[requestIndex] = request;
                continue insertPath;
            }
        }

        if (!pendingRequest) {
            pendingRequest = RequestType.create(this, model, this.total++);
            requests[requestIndex] = pendingRequest;
            participatingRequests[requestCount++] = pendingRequest;
        }

        pendingRequest.insertPath(path, false);
    }

    var pathRequests = [];
    var pathRequestsIndex = -1;

    requestIndex = -1;

    while (++requestIndex < requestCount) {
        request = participatingRequests[requestIndex];
        if (request != null) {
            pathRequests[++pathRequestsIndex] = request;
        }
    }

    return pathRequests;
};

RequestQueue.prototype.mergeJSONGraphs = function mergeJSONGraphs(aggregate, response) {

    var depth = 0;
    var contexts = [];
    var messages = [];
    var keystack = [];
    var latestIndex = aggregate.index;
    var responseIndex = response.index;

    aggregate.index = Math.max(latestIndex, responseIndex);

    contexts[-1] = aggregate.jsonGraph || {};
    messages[-1] = response.jsonGraph || {};

    recursing: while (depth > -1) {

        var context = contexts[depth - 1];
        var message = messages[depth - 1];
        var keys = keystack[depth - 1] || (keystack[depth - 1] = Object.keys(message));

        while (keys.length > 0) {

            var key = keys.pop();

            if (key[0] === prefix) {
                continue;
            }

            if (context.hasOwnProperty(key)) {
                var node = context[key];
                var nodeType = getType(node);
                var messageNode = message[key];
                var messageType = getType(messageNode);
                if (isObject(node) && isObject(messageNode) && !nodeType && !messageType) {
                    contexts[depth] = node;
                    messages[depth] = messageNode;
                    depth += 1;
                    continue recursing;
                } else if (responseIndex > latestIndex) {
                    context[key] = messageNode;
                }
            } else {
                context[key] = message[key];
            }
        }

        depth -= 1;
    }

    return aggregate;
};
/* eslint-enable */

module.exports = RequestQueue;
